{-# LANGUAGE FlexibleContexts #-}

module Fibonacci where

import Control.Monad.State
import Control.Monad.Writer
import Control.Monad.Identity
import AOP
import Data.Map
import MRI


--------------------------------------------------------------------------------
-- Projection functions
--------------------------------------------------------------------------------
proj_S :: Monad m => s -> StateT s m a -> m a
proj_S = flip evalStateT

proj_W :: (Monad m, Monoid w) => WriterT w m a -> m a
proj_W = liftM fst . runWriterT

proj_I = runIdentity

--------------------------------------------------------------------------------
-- Logging and memoization advice
--------------------------------------------------------------------------------
log' :: MonadWriter String m => Advice m a b
log' proceed x = do
  tell "Entering function ..."
  proceed x  

memo :: (MonadState (Map a b) m, Ord a) => Advice m a b
memo proceed x = do
   table <- get
   if member x table  then return (table ! x)
                  else do  y <- proceed x
                           table' <- get
                           put (insert x y table')
                           return y

--------------------------------------------------------------------------------
-- Fibonacci functions in the AOP model
--------------------------------------------------------------------------------

t :: Tag
t = 9

fibAOP :: AOPMonad m => Int -> m Int
fibAOP n = case n of
  0 -> return 1
  1 -> return 1
  _ -> do y <- wapp t fibAOP (n-1)
          x <- wapp t fibAOP (n-2)
          return (x+y)

plainFibAOP :: Monad m => Int -> m Int
plainFibAOP = runAOT [] . (wapp t fibAOP)

logFibAOP :: Monad m => Int -> WriterT String m Int
logFibAOP = runAOT [(pcTag t, log')] . (wapp t fibAOP)

memoFibAOP :: Monad m => Int -> StateT (Map Int Int) m Int
memoFibAOP = runAOT [(pcTag t, memo)] . (wapp t fibAOP)


--------------------------------------------------------------------------------
-- Fibonacci functions in the MRI model
--------------------------------------------------------------------------------

fibMRI :: Monad m => Open (Int -> m Int)
fibMRI this n = case n of
  0 -> return 1
  1 -> return 1
  _ -> do y <- this (n-1)
          x <- this (n-2)
          return (x+y)

plainFibMRI :: Monad m => Int -> m Int
plainFibMRI = new fibMRI

logFibMRI :: Monad m => Int -> WriterT String m Int
logFibMRI = new (log' <@> fibMRI)

memoFibMRI :: Monad m => Int -> StateT (Map Int Int) m Int
memoFibMRI = new (memo <@> fibMRI)
